from .._private._concepts import Event


class ProcedureEvent(Event):
    def __str__(self):
        return ""


class PipelineStartEvent(ProcedureEvent):
    pass


class PipelineStopEvent(ProcedureEvent):
    pass



import asyncio
import time
from socket import gaierror
from threading import Lock

import websockets
from websockets.exceptions import ConnectionClosedError

from ._data_types import SocketConfig
from ... import TriggererEventSource, Event, EventSink, PipelineException
from ..._private._utilities import *


class SocketConnector(TriggererEventSource, EventSink):
    def __init__(self,
                 socket_config: SocketConfig,
                 deserializer: Callable[[str], List[Any]],
                 serializer: Callable[[Event], str] = None,
                 initial_message_generator: Callable[[], List[str]] = lambda: [],
                 responder: Callable[[str], List[str]] = lambda message: [],
                 **kwargs):
        super().__init__(**kwargs)
        self.socket_config: SocketConfig = socket_config
        self.deserializer = deserializer
        self.serializer = serializer
        self.initial_message_generator = initial_message_generator
        self.responder = responder
        self.last_try_for_connect = None
        self.current_reconnect_timeout = self.socket_config.initial_reconnect_timeout
        self.websocket: Optional[websockets.WebSocketClientProtocol] = None
        self.running = False
        self.lock = Lock()

    def drown_event(self, event: Event) -> None:
        if self.serializer is None:
            raise PipelineException("Socket Connector cannot serialize event as it has no serializer")
        serialized_event: str = self.serializer(event)
        SocketConnector._get_event_loop().run_until_complete(self._send_message(serialized_event))

    def get_event_batch(self) -> List[Event]:
        event_string = SocketConnector._get_event_loop().run_until_complete(self._get_message_string())
        if event_string is None:
            return []
        try:
            return self.deserializer(event_string)
        except Exception as e:
            self.logger.exception(f"Failed to deserialize socket message \"{event_string}\": {e.__str__()}")

    async def _send_message(self, message: str):
        websocket = await self._get_websocket()
        if websocket is None:
            raise PipelineException("Cannot send message as because failed to load websocket")
        await websocket.send(message)

    async def _get_message_string(self) -> Optional[str]:
        websocket = await self._get_websocket()
        if websocket is None:
            return None
        try:
            message = await websocket.recv()
            responses: List[str] = self.responder(message)
            for response in responses:
                websocket.send(response)
            return message
        except ConnectionClosedError as e:
            self.logger.error(f"Disconnected from {self.socket_config.server.__str__()}: {str(e)}")
        except Exception as e:
            self.logger.exception(f"Disconnected from {self.socket_config.server.__str__()}: {str(e)}")
        self.websocket = None
        return None

    async def _get_websocket(self):
        self.lock.acquire()
        try:
            if self.websocket is None:
                self.websocket = await self._get_initial_websocket()
            return self.websocket
        finally:
            self.lock.release()

    async def _get_initial_websocket(self) -> Optional[websockets.WebSocketClientProtocol]:
        if self.last_try_for_connect is not None and (
                self.last_try_for_connect + self.current_reconnect_timeout) > time.time():
            return None
        self.last_try_for_connect = time.time()
        self.current_reconnect_timeout = min(self.current_reconnect_timeout * 2,
                                             self.socket_config.max_reconnect_timeout)
        self.logger.debug(f"Trying to connect to {self.socket_config.server.__str__()}")
        try:
            result = await websockets.connect(
                uri=self.socket_config.server.address,
                port=self.socket_config.server.port,
                **self.socket_config.connection_kwargs)
            self.logger.info(f"Connected to {self.socket_config.server.__str__()}")
            self.current_reconnect_timeout = self.socket_config.initial_reconnect_timeout
            for initial_message in self.initial_message_generator():
                await result.send(initial_message)
            return result
        except gaierror as e:
            self.logger.info(
                f"Failed to connect to {self.socket_config.server.__str__()}: {str(e)}. Retrying in {self.current_reconnect_timeout} seconds")
        except Exception as e:
            self.logger.exception(
                f"Failed to connect to {self.socket_config.server.__str__()}: {str(e)}. Retrying in {self.current_reconnect_timeout} seconds")
        return None

    @staticmethod
    def _get_event_loop():
        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(event_loop)
            return event_loop

    def _do_periodic_event_generation(self):
        while self.running:
            try:
                events: List[Event] = self.get_event_batch()
                for event in events:
                    self.event_handler.handle_event(event)
            except Exception as e:
                self.logger.exception("Failed to generate event", e)

    def start_generating(self) -> None:
        self.running = True
        threading.Thread(target=self._do_periodic_event_generation).start()

    def stop_generating(self) -> None:
        self.running = False



from ._private._concepts import Event, EventHandler
from ._private._elements import \
    InternalPipelineElement, \
    EventGenerator, EventListener, \
    EventSource, EventTransformer, EventSink, \
    TriggererEventSource, TriggerableEventSource, TriggerableEventBatchSource
from ._private._pipeline import Pipeline
from ._private._data_types import Server
from ._private._common import PipelineException



from ..._private._utilities import *
from ... import Event
from copy import deepcopy


SIMPLE_CHANGE_INDICATOR = lambda x, y: x != y


class DataStore(ABC):
    @abstractmethod
    def get_value(self, key: str) -> Optional[Any]:
        raise NotImplementedError()

    @abstractmethod
    def set_value(self, key: str, value: Any, **kwargs) -> None:
        raise NotImplementedError()

    @abstractmethod
    def contains(self, key: str) -> bool:
        raise NotImplementedError()

    def __getitem__(self, item: str) -> Optional[Any]:
        return self.get_value(item)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set_value(key, value)

    def __contains__(self, item: str) -> bool:
        return self.contains(item)

    # noinspection PyAttributeOutsideInit,PyProtectedMember
    @synchronized
    def get_branch(self, prefix: str,
                   value_generator: Callable[[str], Any] = None,
                   change_event_handlers: List[Callable[[str, Optional[Any], Any, Optional[Event]], None]] = None,
                   change_indicator: Callable[[Any, Any], bool] = SIMPLE_CHANGE_INDICATOR
                   ):
        if not hasattr(self, '_branches'):
            self._branches: Dict[str, DataStoreBranch] = dict()
        if prefix in self._branches:
            result: DataStoreBranch = self._branches[prefix]
            if value_generator is not None:
                if result._value_generator is not None:
                    raise ValueError(f'Branch with prefix "{prefix}" already has a value generator')
                else:
                    result._value_generator = value_generator
            if change_indicator != SIMPLE_CHANGE_INDICATOR:
                if result._change_indicator != SIMPLE_CHANGE_INDICATOR:
                    raise ValueError(f'Branch with prefix "{prefix}" already has a change indicator')
                else:
                    result._change_indicator = change_indicator
            if change_event_handlers is not None and len(change_event_handlers) > 0:
                if result._change_event_handlers is None:
                    result._change_event_handlers = change_event_handlers
                else:
                    result._change_event_handlers += change_event_handlers
        else:
            result = DataStoreBranch(self, prefix, value_generator, change_event_handlers, change_indicator)
            self._branches[prefix] = result
        return result


class DataStoreBranch(DataStore):
    def __init__(self,
                 parent_data_store: DataStore,
                 prefix: str,
                 value_generator: Callable[[str], Any],
                 change_event_handlers: List[Callable[[str, Optional[Any], Any, Optional[Event]], None]],
                 change_indicator: Callable[[Any, Any], bool]
                 ):
        super().__init__()
        self._parent_data_store = parent_data_store
        self._prefix = prefix
        self._value_generator = value_generator
        self._change_event_handlers = change_event_handlers
        self._change_indicator = change_indicator

    def get_value(self, key: str) -> Optional[Any]:
        result = self._parent_data_store.get_value(self._prefix + key)
        if result is None and self._value_generator is not None:
            result = self._value_generator(key)
            self._parent_data_store.set_value(self._prefix + key, result)
            if (self._change_event_handlers is not None
                    and len(self._change_event_handlers) > 0
                    and self._change_indicator(None, result)):
                for handler in self._change_event_handlers:
                    handler(key, None, result, None)
        return result

    def set_value(self, key: str, value: Any, **kwargs) -> None:
        should_call_handlers = False
        if self._change_event_handlers is not None and len(self._change_event_handlers) > 0:
            previous_value = self.get_value(key)
            should_call_handlers = True
            if not self._change_indicator(previous_value, value):
                return
        self._parent_data_store.set_value(self._prefix + key, value, **kwargs)
        if should_call_handlers:
            for handler in self._change_event_handlers:
                # noinspection PyUnboundLocalVariable
                handler(key, previous_value, value, **kwargs)

    def contains(self, key: str) -> bool:
        return self._parent_data_store.contains(self._prefix + key)


class SimpleRAMDataStore(DataStore):
    def __init__(self):
        super().__init__()
        self._storage: Dict[str, Any] = dict()

    def get_value(self, key: str) -> Optional[Any]:
        return deepcopy(self._storage.get(key, None))

    def set_value(self, key: str, value: Any, **kwargs) -> None:
        self._storage[key] = deepcopy(value)

    def contains(self, key: str) -> bool:
        return key in self._storage
